<?php
/**
 * @file
 *  Drupal State module
 */

class State {

  /**
   * Form expiration in seconds.
   *
   * The value cannot be retrieved from sources like $GLOBALS['conf'], since it's just a hardcoded local value of form_set_cache().
   * Btw; form_set_cache()'s value is the same as DRUPAL_MAXIMUM_TEMP_FILE_AGE.
   *
   * @type integer
   */
  const FORM_EXPIRE = 21600; // ~ 6 hours

  /**
   *  List of name-spaced context types, for setting/getting session variables:
   *  - system: name of a system context
   *  - module: name of a module
   *  - application: name of concept that transcends themes and modules
   *
   * @var array $_contextTypes
   */
  protected static $_contextTypes = array(
    'module' => 'module',
    'application' => 'application',
    'system' => 'system',
  );

  /**
   * @var integer|boolean $_frontendSessionControl
   */
  protected static $_frontendSessionControl = FALSE;

  /**
   * @var integer $_checkSession
   */
  protected static $_checkInterval = 300;

  /**
   * Reload page upon session expire?
   *
   * @var integer $_reloadExpired
   */
  protected static $_reloadExpired = 0;

  /**
   * @var boolean $_noFormExpire
   */
  protected static $_noFormExpire = FALSE;

  /**
   * @var boolean|NULL $_countRequests
   */
  protected static $_countRequests;

  /**
   * @var array $_nonPagePaths
   */
  protected static $_nonPagePaths = array(
      'system/ajax/',
      'file/ajax/',
      'file/progress/',
  );

  /**
   * @var boolean|NULL $_isPageRequest
   */
  protected static $_isPageRequest;

  /**
   * @var integer $_pageRequests
   */
  protected static $_pageRequests = -1;

  /**
   * @var integer $_nonPageRequests
   */
  protected static $_nonPageRequests = -1;

  /**
   * @var integer $_loggedInPageRequests
   */
  protected static $_loggedInPageRequests = -1;

  /**
   * @var integer $_pageResponses
   */
  protected static $_pageResponses = -1;

  /**
   * Nobody but the user module should destroy a session (and that only at request processing start),
   * so we assume that a session will never be destroyed later.
   *
   * @var boolean $_started
   */
  protected static $_started = FALSE;

  /**
   * For completely AJAX-driven pages, which solely uses custom AJAX mechanisms, not Drupal form API's AJAX procedures.
   *
   * Makes frontend session prolongation skip checking for form expiration.
   *
   * Has to be called per page (like in theme hook) because it's unlikely that all forms in a Drupal site are completely AJAX-driven.
   *
   *  Why 'custom AJAX' and not 'Drupal form API's AJAX'?
   *  - because Drupal's conventional form AJAX procedures - probably - are constrained by Drupal form expiration (6 hours)
   *
   * @return void
   */
  public static function unlimitFormExpire() {
    self::$_noFormExpire = TRUE;
  }

  /**
   * Log current user out - if logged in - and/or destroy current visitor's session, if any.
   *
   * @param string $session_id
   *  - default: empty
   *  - non-empty: destroy only if value equals current session id
   * @param string $session_id_hash
   *  - default: empty
   *  - non-empty: destroy only if value equals hash of current session id
   * @return void
   */
  public static function destroy($session_id = '', $session_id_hash = '') {
    if ( (self::$_started || (self::$_started = drupal_session_started()))
        && (!$session_id || $session_id === session_id())
        && (!$session_id_hash || $session_id_hash === hash('sha256', session_id() . variable_get('module_state_salt')))
    ) {
      //  Remove State session related non-httponly cookies, which frontend uses.
      //  The hash of session id, used to check if window/tab's session is current.
      setcookie(
        'module_state_sid', '', time() - 3600,
        $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], FALSE // httponly
      );
      //  The session prolonged cookie, used to get the last time any window/tab made a (prolonging) AJAX request.
      setcookie(
        'module_state_ssprlngd', '', time() - 3600,
        $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], FALSE // httponly
      );

      //  We cannot call user_logout(), because that function performs a redirect, but we do all the same procedures (but drupal_goto()).
      //  And we expect that no function implementing hook_user_logout() performs a redirect;
      //  though the consequenses probably would narrow to the cookies above not being deleted.
      //
      //  global $user;
      //  watchdog('user', 'Session closed for %name.', array('%name' => $user->name));
      //  module_invoke_all('user_logout', $user);
      //  // Destroy the current session, and reset $user to the anonymous user.
      //  session_destroy();
      //  drupal_goto();
      //
      //  user_logout()
      global $user;
      if ($user->uid) {
        watchdog('state', 'Session closed for %name due to session timeout.', array('%name' => $user->name));
        module_invoke_all('user_logout', $user);
      }
      session_destroy();
      //  /user_logout()

      self::$_started = FALSE;
    }
  }

  /**
   * Set session variable, with low risk of collision.
   *
   * Starts drupal session, if no $_SESSION.
   *
   * Cannot set NULL as value.
   * Last non-NULL argument is considered value.
   *
   *  Breaks if something else has set a non-array session variable for one of the name spaces, like:
   *  - $_SESSION['system'] = non-array
   *  - $_SESSION['module'] = non-array
   *  - $_SESSION['application'] = non-array
   *
   * @throws Exception
   *  - always caught and logged
   *  - if bad argument
   *  - if too few non-NULL arguments (requires 3 for name-spaced context types, 2 for other contexts)
   *  - if session started and $_SESSION isnt array
   *  - if $context_type is falsy
   *  - if $context_type is name-spaced and $_SESSION[$context_type] exists and isnt array
   *  - if $context_type is name-spaced and arg $context_name used and is numeric
   * @param string|array $context_type
   *  - string, name-spaced value: system|module|application
   *  - string, non-namespaced value: anything else, like 'theme'
   *  - array: all arguments
   * @param mixed $context_name
   *  - default: empty string
   *  - empty or numeric not allowed if name-spaced $context_type (system|module|application)
   * @param mixed $any_number_of_keys
   *  - default: no argument
   *  - last argument is considered value of the session array being set
   * @return boolean|NULL
   *  - NULL: session not started, and could not be started
   *  - FALSE on error
   */
  public static function set($context_type, $context_name = '', $any_number_of_keys = NULL) {
    return self::_access(10, is_array($context_type) ? $context_type : func_get_args());
  }

  /**
   * Increase session variable, with low risk of collision.
   *
   * Sets the session var using the increment value, if the var doesnt exist.
   *
   * Starts drupal session, if no $_SESSION.
   *
   * Last non-NULL argument is considered the increment value.
   *
   *  Breaks if something else has set a non-array session variable for one of the name spaces, like:
   *  - $_SESSION['system'] = non-array
   *  - $_SESSION['module'] = non-array
   *  - $_SESSION['application'] = non-array
   *
   * @throws Exception
   *  - always caught and logged
   *  - if bad argument
   *  - if too few non-NULL arguments (requires 3 for name-spaced context types, 2 for other contexts)
   *  - if session started and $_SESSION isnt array
   *  - if $context_type is falsy
   *  - if $context_type is name-spaced and $_SESSION[$context_type] exists and isnt array
   *  - if $context_type is name-spaced and arg $context_name used and is numeric
   * @param string|array $context_type
   *  - string, name-spaced value: system|module|application
   *  - string, non-namespaced value: anything else, like 'theme'
   *  - array: all arguments
   * @param mixed $context_name
   *  - default: empty string
   *  - empty or numeric not allowed if name-spaced $context_type (system|module|application)
   * @param mixed $any_number_of_keys
   *  - default: no argument
   *  - last argument is considered the increment value of the session array being set
   * @return number|NULL|boolean
   *  - number: the value after increment
   *  - NULL: session not started, and could not be started
   *  - FALSE on error
   */
  public static function increase($context_type, $context_name = '', $any_number_of_keys = NULL) {
    return self::_access(11, is_array($context_type) ? $context_type : func_get_args());
  }

  /**
   * Get session variable, with low risk of collision.
   *
   *  Breaks if something else has set a non-array session variable for one of the name spaces, like:
   *  - $_SESSION['system'] = non-array
   *  - $_SESSION['module'] = non-array
   *  - $_SESSION['application'] = non-array
   *
   * Asking for key in a key path that isnt array is not considered an error; example:
   * @code
   *   State::get('theme') ~ 'current theme name', swell
   *   State::get('theme', 'whatever') ~ NULL, no error
   * @endcode
   * @throws Exception
   *  - always caught and logged
   *  - if bad argument
   *  - if session started and $_SESSION isnt array
   *  - if $context_type is falsy
   *  - if $context_type is name-spaced and $_SESSION[$context_type] exists and isnt array
   *  - if $context_type is name-spaced and arg $context_name used and is numeric
   * @param string|array $context_type
   *  - string, name-spaced value: system|module|application
   *  - string, non-namespaced value: anything else, like 'theme'
   *  - array: all arguments
   * @param string|integer|NULL $context_name
   *  - default: no argument
   *  - numeric not allowed if name-spaced $context_type (system|module|application)
   * @param string|integer|NULL $any_number_of_keys
   *  - default: no argument
   * @return mixed
   *  - NULL: session not started, or that session array key doesnt exist, or it's value is NULL
   *  - FALSE: error, or that session array value is FALSE
   *  - otherwise the value of that array
   */
  public static function get($context_type, $context_name = NULL, $any_number_of_keys = NULL) {
    return self::_access(0, is_array($context_type) ? $context_type : func_get_args());
  }

  /**
   * Get and unset session variable, with low risk of collision.
   *
   * Acts precisely like State::get(), except removes that session variable.
   *
   * Uses array_splice() - re-indexing the array - if the last argument is an integer (not 'integer'),
   * and that index is placed correctly in the (presumably numerically indexed) array.
   *
   * @see State::get()
   * @param string|array $context_type
   *  - string, name-spaced value: system|module|application
   *  - string, non-namespaced value: anything else, like 'theme'
   *  - array: all arguments
   * @param string|integer|NULL $context_name
   *  - default: no argument
   *  - numeric not allowed if name-spaced $context_type (system|module|application)
   * @param string|integer|NULL $any_number_of_keys
   *  - default: no argument
   * @return mixed
   *  - NULL: session not started, or that session array key doesnt exist, or it's value is NULL
   *  - FALSE: error, or that session array value is FALSE
   *  - otherwise the value of that array
   */
  public static function remove($context_type, $context_name = NULL, $any_number_of_keys = NULL) {
    return self::_access(1, is_array($context_type) ? $context_type : func_get_args());
  }

  /**
   * Get list of registered non-page paths.
   *
   * @param type $builtInsOnly
   *  - default: FALSE
   * @return array
   */
  public static function nonPagePaths($builtInsOnly = FALSE) {
    return $builtInsOnly || !($v = variable_get('module_state_nonpages')) ? self::$_nonPagePaths :
        array_merge(self::$_nonPagePaths, explode("\n", $v));
  }

  /**
   * Not AJAX-, iframe- and other non-page requests.
   *
   * @return boolean|NULL
   *  - NULL: module not configured to detect and count requests
   */
  public static function isPageRequest() {
    return self::$_isPageRequest;
  }

  /**
   * Request number in session (the whole browser session, not only $_SESSION).
   *
   * Only valid (> -1) if State is configured to detect (and count by) request type.
   *
   * 'logged_in' request number is not available when current request is a non-page request (then returns -1).
   *
   * A page redirect counts as a page request.
   *
   * @param string $type
   *  - default: 'page'
   *  - allowed values: page | non_page | all (~ page + non_page) | logged_in (~ page request since user logged in) | list
   * @return integer|array|FALSE
   *  - FALSE: unsupported $type value
   *  - minus 1: module not configured to detect and count requests, or is non-page request and $type is 'logged_in'
   *  - 1: first request is one, not zero
   *  - array: when $type is 'list'; [ 'page': n0, 'non_page': n1, 'logged_in': n2 ]
   */
  public static function requestNumber($type = 'page') {
    switch ($type) {
      case 'page':
        return self::$_pageRequests;
      case 'non_page':
        return self::$_nonPageRequests;
      case 'logged_in':
        return self::$_loggedInPageRequests;
      case 'all':
        return self::$_pageRequests + self::$_nonPageRequests;
      case 'list':
        return array(
          'page' => self::$_pageRequests,
          'non_page' => self::$_nonPageRequests,
          'logged_in' => self::$_loggedInPageRequests,
        );
    }
    return FALSE;
  }

  /**
   * Number of page responses delivered in session (the whole browser session, not only $_SESSION) - until now (current response not counted).
   *
   * Always fairly reliable, and doesnt depend on State being configured to detect (and count by) request type.
   *
   * @return integer
   *  - zero: nono yet
   */
  public static function pageResponses() {
    return (($n = self::$_pageResponses) > -1) ? $n :
        (self::$_pageResponses = isset($_COOKIE['module_state_prespno']) ? (int)$_COOKIE['module_state_prespno'] : 0);
  }

  /**
   * Checks whether frontend session control - checking/prolongation - is turned on and applys to current visitor.
   *
   * Not called in init, because it should only be called on demand (and available to everybody).
   * Will eventually be called during page generation, via the hook_preprocess_html() implementation.
   *
   * @return integer|boolean
   *  - FALSE: session not started yet, frontend checking/prolongation may apply if session gets started later
   *  - 0: none
   *  - 1: checks session/form expiration
   *  - 2: prolongs session, when visitor is active (mousemove etc.)
   *  - 3: prolongs session unconditionally
   */
  public static function frontendSessionControl() {
    if (!self::$_started && !(self::$_started = drupal_session_started())) {
      return FALSE;
    }
    if (($fsc = self::$_frontendSessionControl) || $fsc === 0) {
      return $fsc;
    }
    if (!variable_get('module_state_frntnd') || !($settings = variable_get('module_state_frntndsttngs'))) {
      return (self::$_frontendSessionControl = 0);
    }
    $fsc = 0;
    $everybody = FALSE;
    $user = NULL;
    //  Visitor activity based prolongation overrides unconditional prolongation - except if activity based is * (everybody).
    //  Activity based prolongation.
    if (!empty($settings->prlngactrls) && ($nRoles = count($roles = $settings->prlngactrls))) {
      if ($nRoles === 1 && $roles[0] == '*') {
        $fsc = 2;
        $everybody = TRUE;
      }
      else {
        $user = user_load($GLOBALS['user']->uid);
        for ($i = 0; $i < $nRoles; $i++) {
          if (in_array($roles[$i], $user->roles)) {
            $fsc = 2;
            break;
          }
        }
      }
    }
    //  Unconditional prolongation.
    if ((!$fsc || $everybody) && !empty($settings->prlngrls) && ($nRoles = count($roles = $settings->prlngrls))) {
      if (!$everybody && // If both are * then activity based rules.
          $nRoles === 1 && $roles[0] == '*') {
        $fsc = 3;
      }
      else {
        if (!$user) {
          $user = user_load($GLOBALS['user']->uid);
        }
        for ($i = 0; $i < $nRoles; $i++) {
          if (in_array($roles[$i], $user->roles)) {
            $fsc = 3;
            break;
          }
        }
      }
      //  Unconditional prolongiation is not allowed for user profile editing pages.
      if ($fsc == 3 && count($args = arg()) > 2 && $args[0] == 'user' && preg_match('/^\d+$/', $args[1]) && $args[2] == 'edit') {
        $fsc = $everybody ? 2 : 0;
      }
    }
    if (!$fsc && !empty($settings->chcksss)) {
      $fsc = 1;
    }
    if ($fsc) {
      if (!empty($settings->chckntrvl)) {
        self::$_checkInterval = $settings->chckntrvl;
      }
      if (!empty($settings->sssrld)) {
        self::$_reloadExpired = 1;
      }
    }
    return (self::$_frontendSessionControl = $fsc);
  }

  /**
   * Adds state javascript library, and commands session checking and/or prolongation (if applicable for current visitor).
   *
   * Public because state_preprocess() needs access.
   *
   * @staticvar boolean $called
   * @return void
   */
  public static function _frontend() {
    static $called = FALSE;
    if (!$called) {
      $called = TRUE;
      //  Page response number (how many times we have delivered a page).
      setcookie(
          'module_state_prespno',
          $pageRespNo = isset($_COOKIE['module_state_prespno']) ? ((self::$_pageResponses = (int)$_COOKIE['module_state_prespno']) + 1) : 1, 0,
          $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], TRUE // httponly
      );
      //  Check if not checked yet (or no session).
      if (($fsc = self::$_frontendSessionControl) === FALSE) {
        $fsc = self::frontendSessionControl();
      }
      if (!$fsc) {
        drupal_add_js(
            '(function($) { $(document).bind("ready", function() { State.init(); }); })(jQuery);',
            array('type' => 'inline', 'scope' => 'header', 'preprocess' => FALSE, 'every_page' => FALSE)
        );
        return;
      }
      //  Add libraries.
      drupal_add_library('system', 'cookie'); // jQuery.cookie
      drupal_add_library('system', 'ui.dialog'); // jQuery.ui.dialog, for expiration info to visitor.
      //  Hash of session id must be highly obfuscated, because we set it in non-httponly cookie (session cookie is httponly).
      $sessHash = hash('sha256', session_id() . variable_get('module_state_salt'));
      setcookie(
          'module_state_sid', $sessHash, 0, // for session
          $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], FALSE // not httponly
      );
      $options = 'checkInterval:' . self::$_checkInterval . ', '
          . 'prolong:' . $fsc . ', '
          . 'reloadExpired:' . self::$_reloadExpired . ', '
          . 'localEn:' . ($GLOBALS['language_content']->language != 'en' || !($v = variable_get('module_state_enovrrd')) ? '{}' : $v);
      drupal_add_js(
          '(function($) { $(document).bind("ready", function() { '
              . 'State.init("' . $sessHash . '", ' . $pageRespNo . ', ' . (!self::$_noFormExpire ? state::FORM_EXPIRE : 0)
              . ', { ' . $options . ' }); }); })(jQuery);',
          array('type' => 'inline', 'scope' => 'header', 'preprocess' => FALSE, 'every_page' => FALSE)
      );
    }
  }

  /**
   * Internal method for accessing/modifying session vars.
   *
   * Public because functions state_set()/state_increase()/state_get()/state_remove() must have access.
   * Using this method directly is not recommended.
   *
   * @throws Exception
   *  - always caught and logged
   * @param integer $act
   * @param a $args
   * @return mixed
   */
  public static function _access($act, $args) {
    switch ($act) {
      case 10: // set
      case 11: // increase
        if (!self::$_started && !(self::$_started = drupal_session_started())) {
          drupal_session_start(); // Why doesnt that return boolean?
          if (!(self::$_started = drupal_session_started())) {
            return NULL;
          }
        }
        break;
      default: // 0 ~ get | 1 ~ remove
        if (!isset($_SESSION) && !$_SESSION) {
          return NULL;
        }
    }
    try {
      if (is_array($_SESSION)) {
        $nArgs = count($args);
        //  Remove NULL values, backwards, which may exist $args is all arguments of calling method/function.
        for ($i = $nArgs - 1; $i >= 0; $i--) {
          if ($args[$i] === NULL) {
            unset($args[$i]);
          }
          else {
            break;
          }
        }
        $nameSpaced = isset(self::$_contextTypes[ $args[0] ]);
        switch ($nArgs = count($args)) {
          case 0:
            throw new Exception('No args');
          case 1:
            if ($act >= 10) { // set/increase must receive more than one arg
              throw new Exception('Cannot set/increase value using only one arg');
            }
            break;
          default:
            if ($nameSpaced) { // Namespaced, special requirements.
              if (!$args[1]) {
                throw new Exception('Cannot access namespaced[' . $args[0] . '] vars using falsy 2nd arg[' . $args[1] . ']');
              }
              if (!is_string($args[1])) {
                throw new Exception('Cannot access namespaced[' . $args[0] . '] vars using non-string[' . gettype($args[1]) . '] 2nd arg');
              }
              if ($act >= 10) { // set/increase
                switch ($nArgs) {
                  case 2:
                    if ($act == 10) { // set
                      throw new Exception('Cannot set namespaced[' . join('][', $args) . '] using less than 3 args');
                    }
                    else { // increase
                      throw new Exception('Cannot increase namespaced[' . join('][', $args) . '] using less than 4 args, received 2');
                    }
                  case 3:
                    if ($act == 11) { // increase
                      throw new Exception('Cannot increase namespaced[' . join('][', $args) . '] using less than 4 args');
                    }
                    elseif (!is_array($args[2])) { // set
                      throw new Exception('When setting namespaced[' . $args[0] . '][' . $args[1]
                            . '] using only 3 args, 3rd arg must be array, not [' . gettype($args[2]) . ']');
                    }
                    break;
                  default:
                    if (!is_string($args[2])) {
                      throw new Exception('When setting/increasing namespaced[' . $args[0] . '][' . $args[1]
                            . '] using more than 3 args, 3rd arg must be string, not [' . gettype($args[2]) . ']');
                    }
                }
                if ($act == 11 && !is_numeric($args[ $nArgs - 1 ])) {
                  $v = array_pop($args);
                  throw new Exception('Cannot increase namespaced[' . join('][', $args) . '] using non-numeric value[' . $v . ']');
                }
              }
            }
        }
        $aS =& $_SESSION;
        if ($act < 10) { // get/remove
          for ($i = 0; $i < $nArgs; $i++) {
            if (array_key_exists($k = $args[$i], $aS)) { // Bucket exists.
              //  Not last argument.
              if ($i < $nArgs - 1) {
                if (is_array($aS[$k])) {
                  $aS =& $aS[$k]; // For next iteration.
                  continue;
                }
                elseif (!$i && $nameSpaced) { // $_SESSION[ context type ] is not array
                  throw new Exception('$_SESSION[' . $args[0] . '], type [' . gettype($aS[ $args[0] ]) . '], is not array');
                }
                return NULL;
              }
              elseif (!$i && $nameSpaced && !is_array($aS[$k])) { // $_SESSION[ context type ] is not array
                throw new Exception('$_SESSION[' . $args[0] . '], type [' . gettype($aS[ $args[0] ]) . '], is not array');
              }
              //  Last argument.
              if (!$act) { // get
                return $aS[$k];
              }
              $v = $aS[$k];
              //  Splice if key is integer and placed correctly in the array.
              if (is_int($k) && $k > -1 && count($aS) > $k) {
                $keys = array_keys($aS);
                if ($keys[$k] == $k) {
                  array_splice($aS, $k, 1);
                }
                else {
                  unset($aS[$k]);
                }
              }
              else {
                unset($aS[$k]);
              }
              return $v;
            }
            return NULL;
          }
        }
        else { // set/increase
          $empty = FALSE; // We dont know, check at first iteration.
          for ($i = 0; $i < $nArgs - 1; $i++) { // Last arg is value/increment, skip that.
            $k = $args[$i];
            if ($empty // If parent didnt exist already or current wasnt array.
                || !array_key_exists($k, $aS) // If current doesnt exist.
                || ($i < $nArgs - 2 && !is_array($aS[$k])) // If current not last key, and current not array.
            ) {
              $empty = TRUE;
              $aS[$k] = array();
            }
            if ($i < $nArgs - 2) { // If current not last key, refer for next iteration.
              $aS =& $aS[$k];
            }
          }
          if ($act == 10) { // set
            $aS[$k] = $args[$nArgs - 1];
            return TRUE;
          }
          elseif (array_key_exists($k, $aS) && is_numeric($aS[$k])) {
            return ($aS[$k] += $args[$nArgs - 1]);
          }
          return ($aS[$k] = $args[$nArgs - 1]);
        }
      }
      throw new Exception('$_SESSION, type [' . gettype($_SESSION) . '], is not array');
    }
    catch (Exception $xc) {
      switch ($act) {
        case 0: // get
          $act = 'get';
          break;
        case 1: // remove
          $act = 'remove';
          break;
        case 10: // set
          $act = 'set';
          break;
        case 11: // increase
          $act = 'increase';
          break;
      }
      self::_errorHandler($xc, 'State::' . $act . '()');
    }
    return FALSE;
  }

  /**
   * Public because state_init() needs access.
   *
   * @staticvar boolean $called
   * @param boolean $started
   * @return void
   */
  public static function _initRequestCounting($started) {
    static $called = FALSE;
    if (!$called) {
      $called = TRUE;
      self::$_countRequests = TRUE;
      if ($started) {
        self::$_started = TRUE;
      }
      //  Check if page request.
      $pageRequest = TRUE;
      if ( (!empty($_SERVER['HTTP_X_REQUESTED_WITH']) && strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) == 'xmlhttprequest')) {
        $pageRequest = FALSE;
      }
      elseif (!empty($_GET['q'])) {
        $le = count(
            $a = !($v = variable_get('module_state_nonpages')) ? self::$_nonPagePaths : array_merge(self::$_nonPagePaths, explode("\n", $v))
        );
        for ($i = 0; $i < $le; $i++) {
          if (strpos($_GET['q'], $a[0]) === 0) {
            $pageRequest = FALSE;
            break;
          }
        }
      }
      //  Find non-page request number in cookie.
      $n = !$pageRequest ? 1 : 0;
      if (!empty($_COOKIE['module_state_reqsnonpage'])
          && strlen($v = $_COOKIE['module_state_reqsnonpage']) < 11 && $v > 0 && $v < 2147483647) {
        $n = $v + 1;
      }
      self::$_nonPageRequests = $n;
      if (!$pageRequest) {
        setcookie(
            'module_state_reqsnonpage', $n, 0,
            $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], TRUE // httponly
        );
      }
      //  Find page request number in cookie.
      $n = !$pageRequest ? 0 : 1;
      if (!empty($_COOKIE['module_state_reqspage'])
          && strlen($v = $_COOKIE['module_state_reqspage']) < 11 && $v > 0 && $v < 2147483647) {
        $n = $v + 1;
      }
      self::$_pageRequests = $n;
      //  Non-page request gets out here.
      if (!$pageRequest) {
        return;
      }
      setcookie(
          'module_state_reqspage', $n, 0,
          $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], TRUE // httponly
      );
      //  Find request number since login.
      $n = 0;
      if ($started && user_is_logged_in()) {
        $n = self::increase('module', 'state', 'reqsloggedin', 1);
      }
      self::$_loggedInPageRequests = $n;
    }
  }

  /**
   * Traces and logs via Inspect tracer if permitted, falls back on watchdog (no trace).
   *
   * Displays drupal error message, if error level allows that.
   *
   * @param Exception $xc
   * @param string $message
   *  - default: empty
   * @return void
   */
  protected static function _errorHandler($xc, $message = '') {
    if (!inspect_trace($xc, array('depth' => 1, 'message' => $message, 'severity' => WATCHDOG_ERROR))) {
      watchdog(
        'state',
        '!message', // To prevent error when extracting strings for translation.
        array('!message' => ($message ? ($message . ': ') : '') . $xc->getMessage()), // To prevent error when extracting strings for translation.
        WATCHDOG_ERROR
      );
    }
    if ($GLOBALS['conf']['error_level']) {
      drupal_set_message($xc->getMessage(), 'error');
    }
  }
}