/**
 * @file
 *  Drupal State module
 */

(function($) {

/**
 * Checks that session is alive and current, and optionally prolongs session by making AJAX requests.
 * Singleton, instantiated to itself.
 * @constructor
 * @name State
 * @singleton
 */
var State = function($) {
  /**
   * @ignore
   * @private
   * @type {State}
   */
  var self = this,
  /**
   * @ignore
   * @private
   * @type {boolean|undefined}
   */
  _init,
  /**
   * @ignore
   * @private
   * @type {obj}
   */
  _ = {
    //  {int}
    prolong: 0,
    //  (str|undefined)
    sid: undefined,
    //  (int|undefined)
    pageRespNo: undefined,
    //  {int}
    checkFirst: 60 * 1000, // Dont call in right away (may be bouncing page view), and dont wait too long (prolongs session).
    //  {int}
    checkInterval: 0,
    //  {Date|null}
    lastCheckTime: null,
    //  {bool}
    reloadExpired: false,
    //  {int}
    sessWarnTimeout: 0,
    //  {int}
    sessTimeout: 0,
    //  {Date|null}
    sessExpireTime: null,
    //  {Date|null}
    sessWarnTime: null,
    //  {bool}
    sessCurrent: true,
    //  {bool}
    sessExpired: false,
    //  {Date|null}
    sessProlongedTime: null,
    //  {bool}
    activitySensed: false,
    //  {bool}
    revive: false,
    //  {Date|null}
    formExpireTime: null,
    //  {Date|null}
    formWarnTime: null,
    //  {bool}
    formExpired: false,
    //  {arr}
    errors: [],
    //  {str}
    infoCurrent: "",
    //  {int}
    infoLife: 60000
  },
  /**
   * @private
   * @type {str}
   */
  _basePath = "/",
  /**
   * @private
   * @type {bool}
   */
  _secure,
  /**
   * @private
   * @type {bool|undefined}
   */
  _jqUiDial,
  /**
   * @private
   * @type {obj|undefined}
   */
  _localEn,
  //  Declare private methods, to make IDEs list them
  _local, _o, _check, _ajax, _ajaxCompleted, _url, _redirect, _info;
  /**
   * @ignore
   * @private
   * @param {str} nm
   * @return {str}
   */
  _local = function(nm) {
    var s;
    //  S.... Drupal.t() doesnt use the 'g' flag when replace()'ing, so Drupal.t() replacement is utterly useless - and nowhere to report the bug :-(
    if(!(s = _o(_localEn, nm))) { // English t message overridden?
      switch(nm) {
        case "preExpire":
          s = Drupal.t("This page is about to expire!newline!newlinePlease save now,!newlineif you have made any changes.!newlineOtherwise, please !link_startreload the page!link_end.");
          break;
        case "expired_redirect":
          s = Drupal.t("This page has expired!newline!newlinePlease !link_startproceed to the start page!link_end.");
          break;
        case "expired_reload":
          s = Drupal.t("This page has expired!newline!newlinePlease !link_startreload the page!link_end.");
          break;
        default:
          s = "[LOCAL: " + nm + "]";
      }
    }
    return s.replace(/\!newline/g, "\n");
  };
  /**
   * Object/function property getter, Object.hasOwnproperty() alternative.
   *
   * @param {obj} o
   * @param {str|int} k0
   * @param {str|int} [k1]
   * @return {mixed}
   *  - undefined: o not object, or o doesnt have property k0, or the value of o[k1] is undefined; and the same goes if arg k1 is used
   */
  _o = function(o, k0, k1) {
    var t = typeof o;
    return o && (t === "object" || t === "function") && o.hasOwnProperty(k0) ?
        (k1 === undefined ? o[k0] : (
            (o = o[k0]) && ((t = typeof o) === "object" || t === "function") && o.hasOwnProperty(k1) ? o[k1] : undefined
        ) ) : undefined;
  };
  /**
   * @ignore
   * @private
   * @param {Event|falsy} evt
   * @param {str} action
   * @return void
   */
  _check = function(evt, action) {
    var u, u1, d, to, act = action, ci, f;
    if(!_.sessCurrent && !_.sessExpired && !_.formExpired) { // Sanity check.
      return;
    }
    d = _.lastCheckTime = new Date();
    if(!(_.sessCurrent = (_.sid === $.cookie("module_state_sid")))) {
      _.sessCurrent = false;
      _info(
          "expired",
          _.reloadExpired,
          u = _url(_.reloadExpired)
      );
      _redirect(u, _.infoLife);
      //  Get out.
      return;
    }
    if(action === "getTimeout") {
      //  Set visitor activity listeners, if conditional prolongation.
      if(_.prolong === 2) {
        $(document.documentElement).bind("click mousemove keyup", f = function() {
          _.activitySensed = true;
          if(!_.revive) {
            return;
          }
          _.revive = false;
          if(new Date() < _.formWarnTime) { // The other basic conditions are evaluated in the beginning of _check().
            _check(null, "check");
          }
        });
        $(window).scroll(f);
      }
      //  Do ajax request (fall thru).
    }
    //  We will never get here if the getTimeout request fails, because only successful getTimeout response starts first check timer.
    else if(action === "check") {
      //  Update session expire time, if this or another window/tab has made prolonging request recently.
      if((u = $.cookie("module_state_ssprlngd")) && u) {
        u = u.split(",");
        if(u.length === 2 && u[1] === _.sid && /^\d{13,14}$/.test(u[0]) &&
            (u = parseInt(u[0], 10)) && (u1 = new Date(u)) &&
            u1 > _.sessProlongedTime) {
          _.sessProlongedTime = u1;
          _.sessExpireTime = new Date( u + _.sessTimeout );
          _.sessWarnTime = new Date( u + _.sessWarnTimeout ); // Warn visitor 10% earlier than timeout.
        }
      }
      ci = _.checkInterval;
      if(d >= _.sessExpireTime) { // Destroy session, if expired.
        _.sessExpired = true;
        _info(
            "expired",
            _.reloadExpired,
            u = _url(_.reloadExpired)
        );
        //  Help session garbage collection; destroy the session via backend.
        //  But wait, some other window/tab might be doing some work, which should be allowed to finish before session is destroyed
        to = setTimeout(function() {
          if(_.sid === $.cookie("module_state_sid")) {
            _ajax("destroySession");
          }
        }, ci = Math.floor(ci / 2));
        _redirect(u, ci + _.infoLife); // One minute to let destroySession call sink in.
        //  Get out.
        return;
      }
      else if((u = _.formExpireTime) && d >= u) { // Dont prolong session, if form expired.
        _.formExpired = true;
        _info(
            "expired",
            true, // Always reload when form expired.
            u = _url(true) // Always reload when form expired.
        );
        _redirect(u, _.infoLife);
        return;
      }
      //  Tell visitor if form about to expire (and not telling that at the moment).
      else if((u = _.formWarnTime) && d >= u && _.infoCurrent !== "preExpire_form") {
        _info(
          "preExpire_form",
          true, // Always reload when warning.
          _url(true) // Always reload when warning.
        );
      }
      //  Prolong session?
      act = null;
      switch(_.prolong) {
        case 3: // Unconditional prolongation
          act = "prolong"; // Continue to ajax request
          break;
        case 2: // User activity based prolongation
          if(_.activitySensed) {
            _.activitySensed = false; // Clear for next interval.
            //  If was closing in on session expiration, clear such info.
            if(_.infoCurrent === "preExpire_sess") {
              _info(); // Hide, and clear _.infoCurrent.
            }
            act = "prolong"; // Continue to ajax request
          }
          break;
        // default: 0|1 ~ no prolongation
      }
      //  Regulate check timeout.
      //  Skip prolongation and modify timeout if this or another browser tab recently prolonged session (~ obsolete requests when multiple tabs).
      if((u = _.sessProlongedTime) &&
          //  The 0.8 multiplier is part of fuzzy logic algo, which ensures approximately correct interval, even if more than one browser tab.
          u.getTime() > d.getTime() - (ci * 0.8)) {
        act = "skip"; // Do setTimeout, but skip AJAX call.
        //  Random number between 60% and full checkInterval (part of fuzzy logic algo).
        ci = Math.floor((ci * 0.6) + ( ( Math.random() * ( (ci * 0.4) + 1 ) ) + 1 ) - 1);
      }
      //  If closing in to session or form expire - risk of timeout before next check - make next checkInterval smaller (half).
      if((u = new Date(d.getTime() + ci)) > _.sessExpireTime || u > _.formExpireTime) {
        ci = Math.floor(ci / 2);
      }
      //  Start new check timer
      to = setTimeout(function() {
        _check(null, "check");
      }, ci);
      if (!act) { // No prolongation at all, or no user activity sensed (not relevant when unconditional prolongation, because then no session expire).
        //  Tell visitor if session about to expire (and not telling anything at the moment).
        if(!_.infoCurrent && d >= _.sessWarnTime) {
          _info(
            "preExpire_sess",
            true, // Always reload when warning.
            _url(true) // Always reload when warning.
          );
          if(_.prolong === 2) { // If conditional prolongation, allow user activity to check/prolong session and hide preExpire_sess warning.
            _.revive = true;
          }
        }
        return;
      }
    }
    //  Ping server to keep session alive (or, first time, retrieve session timeout).
    if(act === "prolong" || act === "getTimeout") { // And not 'skip'.
      _ajax(act);
    }
  };
  /**
   * @param {str} act
   * @return {void}
   */
  _ajax = function(act) {
    $.ajax({
      url: "/state/ajax/" + _.pageRespNo,
      type: "POST",
      data: {
        action: act,
        arg: act !== "destroySession" ? null : _.sid
      },
      dataType: "json", // expects json formatted response data
      cache: false,
      /**
        * @return {void}
        * @param {obj} responseData
        *  - (str) action
        *  - (bool) success
        *  - (string) error
        *  - (string) message
        *  - (int|undefined) sessionTimeout
        * @param {str} textStatus
        * @param {obj} jqXHR
        */
      success: function(oResp, textStatus, jqXHR) {
        var u, d, t, st, sw, ci, to, to1;
        if(textStatus === "success" && $.type(oResp) === "object") {
          if(oResp.success) {
            if(oResp.action === "destroySession" ||
                !_.sessCurrent || _.sessExpired) {
              return;
            }
            if(_.sid !== $.cookie("module_state_sid")) {
              _.sessCurrent = false;
              _info(
                  "expired",
                  _.reloadExpired,
                  u = _url(_.reloadExpired)
              );
              _redirect(u, _.infoLife);
              return;
            }
            t = (d = new Date()).getTime();
            switch(oResp.action) {
              case "prolong":
                _.sessExpireTime = new Date( t + _.sessTimeout );
                _.sessWarnTime = new Date( t + _.sessWarnTimeout );
                break;
              case "getTimeout":
                if((st = oResp.sessionTimeout)) {
                  //  Subtract checkInterval from server session timeout, otherwise a window/tab may face destroyed session unknowingly.
                  _.sessTimeout = st = (st * 1000) - (ci = _.checkInterval); // Server sessionTimeout is in seconds, not milliseconds.
                  _.sessExpireTime = new Date( t + st );
                  //  Modify checkInterval if more than 20% of session timeout.
                  if(ci > st * 0.2) {
                    _.checkInterval = ci = Math.floor(st * 0.2);
                  }
                  //  Warn visitor 10% earlier than timeout
                  //  But the time between warning and timeout must be at least checkInterval;
                  //  to raise the probability that the session warning actually gets displayed.
                  if((sw = Math.floor(st * 0.1)) < ci) {
                    sw = ci;
                  }
                  _.sessWarnTime = new Date( t + (_.sessWarnTimeout = st - sw) );
                  //  Start timer.
                  to = setTimeout(function(){
                    _check(null, "check");
                  }, ci);
                  //	Register AJAX complete listener.
                  to1 = setTimeout(function(){
                    $("body").ajaxComplete(_ajaxCompleted);
                  }, 1000);
                }
                break;
              default: // unknown action
                try {
                  throw new Error("Unknown action["+oResp.action+"]")
                }
                catch(er) {
                  inspect.trace(er, "State._check()");
                }
                return;
            }
            _.sessProlongedTime = d;
            $.cookie("module_state_ssprlngd", "" + t + "," + _.sid, { // set last time session prolonged cookie
                path: _basePath,
                secure: _secure
            });
          }
          else if(oResp.error === "no_session") {
            _.sessCurrent = false;
          }
          else {
            inspect(oResp, "State._check()");
          }
        }
        else {
          _.errors.push("response: " + textStatus);
          inspect({
              textStatus: textStatus,
              response_data: oResp
          }, "State._check()");
        }
      },
      error: function(jqXHR, textStatus, errorThrown) {
        _.errors.push("response: " + textStatus);
        inspect({
            textStatus: textStatus,
            errorThrown: errorThrown
        }, "State._check()");
      }
    });
  };
  /**
   * jQuery AJAX handler, which tells State that the current session got prolonged now, due to response of a jQuery AJAX request.
   *
   * @ignore
   * @param {Event} e
   * @param {XMLHttpRequest} xhr
   * @param {obj} settings
   * @return {void}
   */
  _ajaxCompleted = function(e, xhr, settings) {
    var u;
    if (_o(xhr, "status") === 200 &&
        typeof (u = _o(settings, "url")) === "string" && u.indexOf("/") === 0 &&
        u.indexOf("/state/ajax/") !== 0) { // State's own responses are handled in _check
      self.sessionProlonged();
    }
  };
  /**
   * Creates url to either / or current url.
   *
   * Removes trailing hash; # but no value.
   *
   * @param {bool} [reload]
   * @return {str}
   *  - url to /, or current url
   */
  _url = function(reload) {
    var loc = window.location, v, le;
    if(!reload) {
      return loc.protocol + "//" + loc.hostname + "/";
    }
    else {
      le = (v = loc.href).length;
      return (v.charAt(le - 1) !== "#" ? v : v.substr(0, le - 1)); // Get rid of trailing #.
    }
  };
  /**
   * Attempts redirect, delayed.
   *
   * @ignore
   * @private
   * @param {str} url
   * @param {int} [delay]
   *  - default: 20 milliseconds
   * @return {void}
   */
  _redirect = function(url, delay) {
    var to = setTimeout(function(){
      window.location.href = url;
    }, delay || 20)
  };
  /**
   * Tell visitor something.
   *
   * @ignore
   * @private
   * @param {str|falsy} type
   *  - if falsy, clears/hides info
   *  - string: preExpire|expired
   * @param {bool} [reload]
   *  - default: falsy
   * @param {str|falsy} [url]
   *  - default: empty
   * @return {void}
   */
  _info = function(type, reload, url) {
    var ms, a, ttl, dial, to;
    _.infoCurrent = type || "";
    if(type) {
      ms = _local((type === "expired" ? (type + (!reload ? "_redirect" : "_reload")) : "preExpire")).
          replace(/\!link_start/, "<a href=\"" + (url || "") + "\">").replace(/\!link_end/, "</a>");
    }
    if(_jqUiDial) {
      if(type) {
        a = ms.split(/\n\n/);
        ttl = a[0].replace(/\n/g, "").replace(/\.$/, "");
        a.splice(0, 1);
        ms = '<p>' + a.join("</p><p>").replace(/\n/g, "<br/>") + "</p>";
      }
      if(!(dial = document.getElementById("module_state_info"))) {
        if(!type) {
          return;
        }
        //  create dialog ----------------------------------------
        if(!document.getElementById("module_state_info_container")) {
          $(document.body).append("<div id=\"module_state_info\">" + ms + "</div>");
        }
        dial = document.getElementById("module_state_info");
        //  instantiate dialog -----------------------------------
        $(dial).dialog({
          stack: true,
          modal: true,
          autoOpen: false,
          closeOnEscape: type !== "expired", // Prohibit close if expired.
          resizable: false,
          position: ["center","center"],
          title : ttl,
          buttons: [
            {
              text: Drupal.t("Ok"),
              click: type !== "expired" ? function(){$(this).dialog("close");} : function(){_redirect(_url(reload));}
            }
          ],
          beforeClose: function() {
            _.infoCurrent = "";
          }
        });
        //  Add our class on dialog's outer container, and make if position:fixed.
        $(dial.parentNode).css("position", "fixed").addClass("module-state-dialog-container");
      }
      else if(!type) {
        $(dial).dialog("close");
        return;
      }
      else {
        $(dial).dialog("option", "title", ttl);
        $(dial).html(ms);
      }
      if(type === "expired") {
        $("a.ui-dialog-titlebar-close", "div.module-state-dialog-container").unbind().hide();
        $(dial).dialog("option", "buttons", [
          {
            text: Drupal.t("Ok"),
            click: function(){_redirect(_url(reload));}
          }
        ]);
      }
      //  Show the dialog; delayed, otherwise modal overlay wont cover the whole window.
      to = setTimeout(function(){
        $(dial).dialog("open");
      }, 5000); // Safe-guarding against very slow rendering.
    }
    //  Using alert() may not work well, because alert() may block timers (and it does prevent redirect in IE):
    else if(type) {
      alert(ms.replace(/\!link_[a-z]+/g, ""));
    }
    //  Cant close alert() programmatically; only user can do that.
  };

  /**
   * Setup session checking/prolongation.
   *
   *  Options:
   *  - (int) checkInterval (default 300), in seconds
   *  - (int) prolong (default 0), 0|1 ~ no | 2 ~ visitor activity conditioned | 3 ~ unconditional
   *  - (bool) reloadExpired (default false ~ redirect to front page on expired)
   *  - (obj) localEn (default empty)
   * @function
   * @name State.init
   * @param {str} sid
   *  - hash of session id (at page load)
   * @param {int} pageRespNo
   *  - page response number ~ how many times backend State has positively detected that a (any) page was delivered
   * @param {int} formExpire
   *  - in seconds
   *  - zero: turns form expiration checks off
   * @param {obj} [options]
   * @return {void}
   */
  this.init = function(sid, pageRespNo, formExpire, options) {
    var fx = formExpire, o = options || {}, u, v, t = new Date().getTime();
    if(_init) {
      return;
    }
    _init = true;
    if(!sid || !pageRespNo || !fx) {
      return;
    }
    _.sid = sid;
    _.pageRespNo = pageRespNo;
    _.formExpireTime = fx ? new Date(t + (fx * 1000)) : null;
    _.formWarnTime = fx ? new Date(t + (fx * 900)) : null; // Warn visitor 10% earlier than expiration.
    _.checkInterval = ((u = _o(o, "checkInterval")) ? u : 300) * 1000;
    _.prolong = (u = _o(o, "prolong")) > 1 ? u : 0;
    _.reloadExpired = _o(o, "reloadExpired") ? 1 : 0;
    _localEn = (u = _o(o, "localEn")) ? u : {};
    if((u = _o(Drupal, "settings", "basePath"))) {
      _basePath = u;
    }
    _secure = window.location.protocol === "https:";
    _jqUiDial = typeof _o($, "ui", "dialog") === "function";
    //  Start timer
    t = setTimeout(function(){
      _check(null, "getTimeout");
    }, _.checkFirst = (u = _.checkFirst) < (v = _.checkInterval) ? u : v); // First interval should not be larger than general interval.
  };
  /**
   * Page load number; how many times backend State has positively detected that a (any) page was delivered.
   *
   * @function
   * @name State.pageResponseNumber
   * @return {int|undefined}
   *  - undefined: State not initialized, or session checking/prolongation
   */
  this.pageResponseNumber = function() {
    return _.pageRespNo;
  };
  /**
   * Check whether session seems alive and current, and form hasnt expired.
   *
   * @function
   * @name State.alive
   * @return {bool|undefined}
   *  - undefined: session checking/prolongation not set up
   */
  this.alive = function() {
    return !_.sid ? undefined : (
        (!_.sessCurrent || _.sessExpired || _.formExpired) ? false :
            (_.sessCurrent = (_.sid === $.cookie("module_state_sid")))
    );
  };
  /**
   * Tell State that the current session got prolonged now, due to a (non jQuery) AJAX or iframe request.
   *
   * There is no need to call this upon response of a jQuery AJAX request; State registers all jQuery AJAX responses.
   * @function
   * @name State.sessionProlonged
   * @param {Date|falsy} [date]
   *  - default: now
   * @return {bool}
   *  - false: session or form has expired, or the session related to current page load is not current anymore, or State not initialized
   */
  this.sessionProlonged = function(date) {
    if(this.alive()) {
      $.cookie("module_state_ssprlngd", "" + (_.sessProlongedTime = date || new Date()).getTime() + "," + _.sid, {
          path: _basePath,
          secure: _secure
      });
      return true;
    }
    return false;
  };
  /**
   * Inspect inner state properties, if module Inspect exists.
   * @function
   * @name State.inspect
   * @return {void}
   */
  this.inspect = function() {
    if(typeof window.inspect === "function" && inspect.tcepsnI) {
      inspect(_, "State");
    }
  };
  /**
   * Test info (dialog) method.
   *
   * @example
State.testInfo("preExpire");
setTimeout(function(){ State.testInfo("expired"); }, 10000);
setTimeout(function(){ State.testInfo("expired", true); }, 20000);
   * @function
   * @name State.testInfo
   * @param {str|falsy} type
   *  - if falsy, clears/hides info
   *  - string: preExpire|expired
   * @param {bool} [reload]
   *  - default: falsy
   * @param {str|falsy} [url]
   *  - default: empty
   * @return {void}
   */
  this.testInfo = function(type, reload, url) {
    var s = _.infoCurrent;
    _info(type, reload, url);
    _.infoCurrent = s;
  };
};

window.State = new State($);

})(jQuery);